/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2022 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.18.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose: Initializes and handles the hardware for embOS
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"
#include "system_gd32vf103.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ    (SystemCoreClock / 4)
#define OS_TICK_FREQ     (1000u)
#define OS_INT_FREQ      (OS_TICK_FREQ)

#define OS_TIMER_RELOAD  (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_DISABLED
#endif

/*********************************************************************
*
*       Device specific SFRs
*/
//
//  Machine timer registers
//
#define MTIME            (*(volatile OS_U64*)(0xD1000000u))  // Used to generate OS tick, Timer counter register
#define MTIMECMP         (*(volatile OS_U64*)(0xD1000008u))  // Used to generate OS tick, Timer compare register
//
// Extended Core-local Interrupt Controller
//
#define ECLIC_BASE_ADDR  (0xD2000000u)

/*********************************************************************
*
*       Function prototypes
*
**********************************************************************
*/
void ISR_M_Software(void);
void ISR_M_Timer   (void);

/*********************************************************************
*
*       Externals
*
**********************************************************************
*/
extern void* vectors;
extern void* trap_entry;

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  unsigned int Result;
  OS_I64       Diff;
  OS_U64       Counter;
  OS_U64       Compare;

  Compare = MTIMECMP;
  Counter = MTIME;
  Diff    = (OS_I64)(Counter - Compare);
  if (Diff < 0) {
    Result = (unsigned int)(OS_TIMER_RELOAD + Diff);
  } else {
    Result = (unsigned int)(Diff);
  }
  return Result;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  unsigned int IsPending;
  OS_U64       Counter;
  OS_U64       Compare;

  Compare = MTIMECMP;
  Counter = MTIME;
  if (Compare <= Counter) {
    IsPending = 1u;
  } else {
    IsPending = 0u;
  }
  return IsPending;
}

/*********************************************************************
*
*       _ExceptionHandler()
*
*  Function description
*    This routine may be used to investigate synchronous traps.
*
*  Additional information
*    _ExceptionHandler() is called when a synchronous trap has occurred.
*/
static void _ExceptionHandler(OS_REG_TYPE mcause, OS_REG_TYPE mepc) {
  OS_USE_PARA(mepc);                    // May be used for further investigation.

#if (OS_DEBUG != 0)
  switch (mcause & ECLIC_EXC_CAUSE) {  // Detect specific exception cause:
  case  0: while (1);                  // - Instruction address misaligned.
  case  1: while (1);                  // - Instruction access fault.
  case  2: while (1);                  // - Illegal instruction.
  case  3: while (1);                  // - Breakpoint.
  case  4: while (1);                  // - Load address misaligned.
  case  5: while (1);                  // - Load access fault.
  case  6: while (1);                  // - Store/AMO address misaligned.
  case  7: while (1);                  // - Store/AMO access fault.
  case  8: while (1);                  // - Environment call from U-mode.
  case  9: while (1);                  // - Environment call from S-mode.
  case 11: while (1);                  // - Environment call from M-mode.
  case 12: while (1);                  // - Instruction page fault.
  case 13: while (1);                  // - Load page fault.
  case 15: while (1);                  // - Store/AMO page fault.
  default: while (1);                  // - Reserved. This should never happen.
  }
#else
  OS_USE_PARA(mcause);
  while (1);  // Halt execution
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       ISR_M_Software()
*
*  Function description
*    This routine does not serve any specific purpose, but is included
*    for demonstration only.
*
*  Additional information
*    ISR_M_Software() is called when the Machine Software interrupt is pending.
*    Machine Software interrupt becomes pending when bit 3 (MSIP) of the
*    Machine Interrupt Pending Register (MIP) is set.
*/
void ISR_M_Software(void) {
  OS_INT_Enter();
  //
  // Perform any functionality here.
  //
  OS_INT_Leave();
}

/*********************************************************************
*
*       ISR_M_Timer()
*
*  Function description
*    This routine performs the embOS tick handling.
*
*  Additional information
*    ISR_M_Timer() is called when the Machine Timer interrupt is pending.
*    Machine Timer Interrupt becomes pending when (MTIMECMP >= MTIME).
*/
void ISR_M_Timer(void) {
  OS_U64 Compare;

  OS_INT_Enter();
  Compare = MTIMECMP;  // Read timer compare value
  do {                 // We might have to perform numerous ticks if (machine timer) interrupt has been disabled for extended periods
    OS_TICK_HandleEx();
    Compare += OS_TIMER_RELOAD;
  } while (Compare <= MTIME);
  MTIMECMP = Compare;  // Eventually, write new compare value. Implicitly clears MTIP bit.
  OS_INT_Leave();
}

/*********************************************************************
*
*       OS_TrapHandler()
*
*  Function description
*    This routine is used to handle traps routed to the mtvec vector.
*    It checks whether the trap was caused by an exception or an
*    interrupt and decides how to continue trap handling.
*
*  Additional information
*    Handling of traps highly depends on the implemented interrupt
*    controller and the used mode. Thus, the implementation of the
*    OS_TrapHandler() might me different for other RISC-V devices.
*
*    OS_TrapHandler() forwards exceptions to _ExceptionHandler().
*/
OS_REG_TYPE OS_TrapHandler(OS_REG_TYPE mcause, OS_REG_TYPE mepc) {
  if (mcause & MCAUSE_INT) {
    //
    // This case should never happen as ECLIC is set up to
    // route exceptions and interrupts to separate trap vectors.
    //
    while (1) {
      ;
    }
  } else {
    //
    // Caused by synchronous trap: call fault handler.
    //
    _ExceptionHandler(mcause, mepc);
  }
  return mepc;
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_INT_IncDI();
  //
  // Initialize interrupt handling (ECLIC)
  //
  OS_ECLIC_Init((void*)ECLIC_BASE_ADDR, (void*)&vectors, (void*)&trap_entry);
  //
  // Install and enable Machine Timer Interrupt
  //
  OS_ECLIC_SetIntPriority(IRQ_M_TIMER, 0);
  OS_ECLIC_EnableInt(IRQ_M_TIMER);
  //
  // Set-up the OS tick interrupt timer
  //
  MTIME    = 0;                // Configure counter register
  MTIMECMP = OS_TIMER_RELOAD;  // Configure compare register
  //
  // Inform embOS about the timer settings
  //
  {
    OS_SYSTIMER_CONFIG _SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&_SysTimerConfig);
  }
  OS_TICK_Config(OS_TICK_FREQ, OS_INT_FREQ);
  //
  // Configure and initialize SEGGER SystemView
  //
#if (OS_SUPPORT_TRACE_API != 0)
  SEGGER_SYSVIEW_Conf();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*    It may be used to e.g. enter a low power mode of the device.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*    The idle loop can be exited only when an embOS interrupt causes
*    a context switch (e.g. after expiration of a task timeout),
*    hence embOS interrupts must not permanently be disabled.
*/
void OS_Idle(void) {            // Idle loop: No task is ready to execute
  while (1) {                   // Nothing to do ... wait for interrupt
    #if (OS_DEBUG == 0)
      //
      // When uncommenting this line, please be aware device
      // specific issues could occur.
      // Therefore, we do not call __WFI() by default.
      //
      //__asm volatile("wfi");  // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if   (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKMEM_SendChar(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USE_PARA(c);          // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*************************** End of file ****************************/
