/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2022 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.18.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose   : Initializes and handles the hardware for embOS

Literature: [1] SiFive E3 Coreplex Series Manual Version 1.2
                \\fileserver\Techinfo\Company\SiFive\MCU\E3-Coreplex-v1.2.pdf
            [3] SiFive E31/E51 Coreplex FPGA Eval Kit User Guide
                \\fileserver\Techinfo\Company\SiFive\MCU\SiFiveE3xxCoreplexFPGAEvalKit_v1p0.pdf
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Interrupt handling configuration
*/
#ifndef   USE_VECTORED_INT_MODE
  #define USE_VECTORED_INT_MODE  (1)  // Configures ISR handling for vectored mode instead of direct mode.
#endif

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ    (32768u)
#define OS_TICK_FREQ     ( 1000u)
#define OS_INT_FREQ      ( 1024u)

#define OS_TIMER_RELOAD  (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_DISABLED
#endif

/*********************************************************************
*
*       Device specific SFRs
*/
//
//  Machine timer registers
//
#define MTIME                 (*(volatile OS_U64*)(0x0200BFF8u))  // Used to generate OS tick, Timer counter register
#define MTIMECMP              (*(volatile OS_U64*)(0x02004000u))  // Used to generate OS tick, Timer compare register
//
// Core-local interrupts
//
#define NUM_LOCAL_INTERRUPTS  (32)
//
// Platform-Level Interrupt Controller
//
#define PLIC_BASE_ADDR        (0x0C000000u)
#define PLIC_NUM_INTERRUPTS   (26)
#define PLIC_NUM_INT_PRIOS    ( 7)

/*********************************************************************
*
*       Function prototypes
*
**********************************************************************
*/
void ISR_M_Software(void);
void ISR_M_Timer   (void);
void ISR_M_External(void);
void ISR_Local4    (void);
void ISR_External2 (void);

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
#if (USE_VECTORED_INT_MODE == 0)
static OS_IRQ_HANDLER* _apfLocalIRQHandler[NUM_LOCAL_INTERRUPTS] __attribute__ ((aligned (4), section (".non_init")));
#endif
static OS_IRQ_HANDLER* _apfGlobalIRQHandler[PLIC_NUM_INTERRUPTS] __attribute__ ((aligned (4), section (".non_init")));

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  unsigned int Result;
  OS_I64       Diff;
  OS_U64       Counter;
  OS_U64       Compare;

  Compare = MTIMECMP;
  Counter = MTIME;
  Diff    = (OS_I64)(Counter - Compare);
  if (Diff < 0) {
    Result = (unsigned int)(OS_TIMER_RELOAD + Diff);
  } else {
    Result = (unsigned int)(Diff);
  }
  return Result;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return ((unsigned int)OS_CLINT_GetIntPending(IRQ_M_TIMER));
}

/*********************************************************************
*
*       _ExceptionHandler()
*
*  Function description
*    This routine may be used to investigate synchronous traps.
*
*  Additional information
*    _ExceptionHandler() is called when a synchronous trap has occurred.
*/
static void _ExceptionHandler(OS_REG_TYPE mcause, OS_REG_TYPE mepc) {
  OS_USE_PARA(mepc);                // May be used for further investigation.

#if (OS_DEBUG != 0)
  switch (mcause & MCAUSE_CAUSE) {  // Detect specific exception cause:
  case  0: while (1);               // - Instruction address misaligned.
  case  1: while (1);               // - Instruction access fault.
  case  2: while (1);               // - Illegal instruction.
  case  3: while (1);               // - Breakpoint.
  case  4: while (1);               // - Load address misaligned.
  case  5: while (1);               // - Load access fault.
  case  6: while (1);               // - Store/AMO address misaligned.
  case  7: while (1);               // - Store/AMO access fault.
  case  8: while (1);               // - Environment call from U-mode.
  case  9: while (1);               // - Environment call from S-mode.
  case 11: while (1);               // - Environment call from M-mode.
  case 12: while (1);               // - Instruction page fault.
  case 13: while (1);               // - Load page fault.
  case 15: while (1);               // - Store/AMO page fault.
  default: while (1);               // - Reserved. This should never happen.
  }
#else
  OS_USE_PARA(mcause);
  while (1);  // Halt execution
#endif
}

/*********************************************************************
*
*       _ISR_NotInstalled()
*
*  Function description
*    This routine may be used to detect non-installed interrupts.
*
*  Additional information
*    _ISR_NotInstalled() is called when an interrupt is pending for
*    which no specific interrupt handler was previously installed.
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
static void _ISR_NotInstalled(void) {
  volatile int Dummy;

  Dummy = 1;
  while (Dummy > 0) {
    //
    // You may set a breakpoint here to detect Interrupts for which no ISR was registered
    //
  }
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       ISR_M_Software()
*
*  Function description
*    This routine does not serve any specific purpose, but is included
*    for demonstration only.
*
*  Additional information
*    ISR_M_Software() is called when the Machine Software interrupt is pending.
*    Machine Software interrupt becomes pending when bit 3 (MSIP) of the
*    Machine Interrupt Pending Register (MIP) is set.
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void ISR_M_Software(void) {
  OS_INT_Enter();
  OS_CLINT_ClearIntPending(IRQ_M_SOFTWARE);  // Explicitly clear MSIP bit.
  //
  // Perform any functionality here.
  //
  OS_INT_Leave();
}

/*********************************************************************
*
*       ISR_M_Timer()
*
*  Function description
*    This routine performs the embOS tick handling.
*
*  Additional information
*    ISR_M_Timer() is called when the Machine Timer interrupt is pending.
*    Machine Timer Interrupt becomes pending when (MTIMECMP >= MTIME).
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void ISR_M_Timer(void) {
  OS_U64 Compare;

  OS_INT_Enter();
  Compare = MTIMECMP;  // Read timer compare value
  do {                 // We might have to perform numerous ticks if (machine timer) interrupt has been disabled for extended periods
    OS_TICK_HandleEx();
    Compare += OS_TIMER_RELOAD;
  } while (Compare <= MTIME);
  MTIMECMP = Compare;  // Eventually, write new compare value. Implicitly clears MTIP bit.
  OS_INT_Leave();
}

/*********************************************************************
*
*       ISR_M_External()
*
*  Function description
*    This routine detects the specific reason for the global IRQ and
*    calls the respective interrupt service routine that was previously
*    installed by OS_PLIC_InstallISR().
*
*  Additional information
*    ISR_M_External() is called when the Machine External interrupt is
*    pending. Machine External interrupt becomes pending when it is
*    asserted by the external Platform-Level interrupt controller (PLIC).
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*/
void ISR_M_External(void) {
  OS_U32 IRQIndex;

  OS_INT_Enter();
  IRQIndex = OS_PLIC_ClaimInt();       // Claim highest-priority pending global interrupt. Implicitly clears MEIP if no further global interrupts are pending.
  if (IRQIndex != 0u) {                // "0" indicates no IRQ was pending at the time of the latest claim.
    _apfGlobalIRQHandler[IRQIndex]();  // Call appropriate handler.
    OS_PLIC_CompleteInt(IRQIndex);     // Signal interrupt completion to PLIC.
  }
  OS_INT_Leave();
}

/*********************************************************************
*
*       ISR_External2()
*
*  Function description
*    This routine does not serve any specific purpose, but is included
*    for demonstration only.
*
*  Additional information
*    ISR_External2() is called by ISR_M_External() when serving
*    the Machine External interrupt and reading external ID 2 from the
*    PLIC.
*    External ID 2 is issued by the PLIC when SW0 is activated on the
*    Arty board.
*
*    Be aware that the pending condition is not cleared by software (it is
*    asserted as long as SW0 is active). Nested interrupts can cause the
*    interrupt routine to nest itself repeatedly, likely causing a stack
*    overflow. However, external ISRs are routed through the PLIC and
*    therefore add minimal latency to ISR procession, which avoids nesting
*    in most cases. It may also be avoided by not allowing nested interrupts
*    at all.
*/
void ISR_External2(void) {
  //
  // Perform any functionality here.
  //
}

/*********************************************************************
*
*       ISR_Local4()
*
*  Function description
*    This routine does not serve any specific purpose, but is included
*    for demonstration only.
*
*  Additional information
*    ISR_Local4() is called when machine local interrupt 20 is pending.
*    Machine local interrupt 20 becomes pending when BTN0 is pressed on the
*    Arty board.
*
*    In direct mode, this routine is called by OS_TrapHandler().
*    In vectored mode, this routine is called from the ROM vector table.
*
*    Be aware that the pending condition is not cleared by software (it is
*    asserted as long as BTN0 is pressed). Nested interrupts can cause the
*    interrupt routine to nest itself repeatedly, likely causing a stack
*    overflow. Since local interrupts are taken without latency, nesting
*    should be avoided by not allowing nested interrupts at all.
*/
void ISR_Local4(void) {
  OS_INT_Enter();
  //
  // Perform any functionality here.
  //
  OS_INT_Leave();
}

/*********************************************************************
*
*       OS_TrapHandler()
*
*  Function description
*    This routine is used to handle traps routed to the mtvec vector.
*    It checks whether the trap was caused by an exception or an
*    interrupt and decides how to continue trap handling.
*
*  Additional information
*    Handling of traps highly depends on the implemented interrupt
*    controller and the used mode. Thus, the implementation of the
*    OS_TrapHandler() might me different for other RISC-V devices.
*
*    OS_TrapHandler() forwards exceptions to _ExceptionHandler().
*/
OS_REG_TYPE OS_TrapHandler(OS_REG_TYPE mcause, OS_REG_TYPE mepc) {
  if (mcause & MCAUSE_INT) {
#if (USE_VECTORED_INT_MODE == 0)
    //
    // Caused by interrupt: call appropriate high-level handler.
    //
    _apfLocalIRQHandler[mcause & MCAUSE_CAUSE]();
#else
    //
    // In vectored mode, user mode software interrupt ambiguously shares a vector with synchronous exceptions.
    // If user mode software interrupt is to be used by the application, its handler would be called here.
    //
    _ISR_NotInstalled();
#endif
  } else {
    //
    // Caused by synchronous trap: call fault handler.
    //
    _ExceptionHandler(mcause, mepc);
  }
  return mepc;
}

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
static const OS_SYSTIMER_CONFIG _SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
void OS_InitHW(void) {
  OS_INT_IncDI();
  //
  // Initialize core-local interrupt handling
  //
#if (USE_VECTORED_INT_MODE == 0)
  OS_CLINT_Init(NUM_LOCAL_INTERRUPTS, _apfLocalIRQHandler);  // Implicitly disables all sources
  for (CLINT_IRQn i = IRQ_U_SOFTWARE; i < NUM_LOCAL_INTERRUPTS; i++) {
    (void)OS_CLINT_InstallISR(i, _ISR_NotInstalled);         // Install dummy handler (allows to omit NULL-pointer checks in OS_TrapHandler())
  }
  OS_CLINT_SetDirectMode();                                  // Replace the default trap_entry which was set during startup and set mode to direct
#else
  OS_CLINT_Init(NUM_LOCAL_INTERRUPTS, NULL);                 // Implicitly disables all sources
  OS_CLINT_SetVectoredMode();                                // Replace the default trap_entry which was set during startup and set mode to vectored
#endif
  //
  // Install and enable Machine Timer Interrupt
  //
#if (USE_VECTORED_INT_MODE == 0)
  (void)OS_CLINT_InstallISR(IRQ_M_TIMER, ISR_M_Timer);
#endif
  OS_CLINT_EnableInt(IRQ_M_TIMER);
  //
  // Install and enable Machine External Interrupt
  //
#if (USE_VECTORED_INT_MODE == 0)
  (void)OS_CLINT_InstallISR(IRQ_M_EXTERNAL, ISR_M_External);
#endif
  OS_CLINT_EnableInt(IRQ_M_EXTERNAL);
  //
  // Install and enable Machine Software Interrupt
  //
#if (USE_VECTORED_INT_MODE == 0)
  (void)OS_CLINT_InstallISR(IRQ_M_SOFTWARE, ISR_M_Software);
#endif
  OS_CLINT_EnableInt(IRQ_M_SOFTWARE);
  //
  // Initialize global interrupt handling (PLIC)
  //
  OS_PLIC_Init(PLIC_BASE_ADDR, PLIC_NUM_INTERRUPTS, PLIC_NUM_INT_PRIOS, _apfGlobalIRQHandler);  // Implicitly disables all sources
  for (int i = 0u; i < PLIC_NUM_INTERRUPTS; i++) {
    (void)OS_PLIC_InstallISR(i, _ISR_NotInstalled);          // Install dummy handler (allows to omit NULL-pointer checks in ISR_M_External())
  }
  //
  // Install and enable a global interrupt
  //
  (void)OS_PLIC_InstallISR(2u, ISR_External2);               // Install handler for platform-specific core-local interrupt ID 2
  OS_PLIC_SetIntPriority(2u, 1u);                            // Set priority for this external interrupt to lowest priority, i.e. "1"
  OS_PLIC_EnableInt(2u);                                     // Enable this interrupt in the PLIC
  //
  // Install and enable a platform-specific core-local Interrupt
  //
#if (USE_VECTORED_INT_MODE == 0)
  (void)OS_CLINT_InstallISR(IRQ_LOCAL4, ISR_Local4);
#endif
  OS_CLINT_EnableInt(IRQ_LOCAL4);
  //
  // Set-up the OS tick interrupt timer
  //
  MTIME    = 0;                                              // Configure counter register
  MTIMECMP = OS_TIMER_RELOAD;                                // Configure compare register
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&_SysTimerConfig);
  OS_TICK_Config(OS_TICK_FREQ, OS_INT_FREQ);
  //
  // Configure and initialize SEGGER SystemView
  //
#if (OS_SUPPORT_TRACE_API != 0)
  SEGGER_SYSVIEW_Conf();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*    It may be used to e.g. enter a low power mode of the device.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*    The idle loop can be exited only when an embOS interrupt causes
*    a context switch (e.g. after expiration of a task timeout),
*    hence embOS interrupts must not permanently be disabled.
*/
void OS_Idle(void) {            // Idle loop: No task is ready to execute
  while (1) {                   // Nothing to do ... wait for interrupt
    #if (OS_DEBUG == 0)
      //
      // When uncommenting this line, please be aware device
      // specific issues could occur.
      // Therefore, we do not call __WFI() by default.
      //
      //__asm volatile("wfi");  // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if   (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKMEM_SendChar(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USE_PARA(c);          // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*************************** End of file ****************************/
