/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2022 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.18.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File    : entry.s
Purpose : Entry point and vector table for the SiFive E31.

Additional information:
  [1] The RISC-V Instruction Set Manual Volume II: Privileged Architecture
      ...
      3.1.7 Machine Trap-Vector Base-Address Register (mtvec)
        ...
        When vectored interrupts are enabled, interrupt cause 0, which corresponds to user-mode software interrupts,
        are vectored to the same location as synchronous exceptions. This ambiguity does not arise in practice,
        since user-mode software interrupts are either disabled or delegated to user mode.
        ...
*/

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

#if (__riscv_xlen == 64)
  #define STORE     sd
  #define LOAD      ld
  #define REGBYTES   8
#else
  #define STORE     sw
  #define LOAD      lw
  #define REGBYTES   4
#endif

/*********************************************************************
*
*       MACROS
*
**********************************************************************
*/

/*********************************************************************
*
*      TRAP_ENTRY
*/
TRAP_ENTRY MACRO
        addi    sp, sp, -20 * REGBYTES
        //
        // Store caller save registers on the stack.
        // Callee save registers will be saved by the respective ISR function.
        //
        STORE   ra,  0 * REGBYTES(sp)
        STORE   t0,  1 * REGBYTES(sp)
        STORE   t1,  2 * REGBYTES(sp)
        STORE   t2,  3 * REGBYTES(sp)
        STORE   a0,  4 * REGBYTES(sp)
        STORE   a1,  5 * REGBYTES(sp)
        STORE   a2,  6 * REGBYTES(sp)
        STORE   a3,  7 * REGBYTES(sp)
        STORE   a4,  8 * REGBYTES(sp)
        STORE   a5,  9 * REGBYTES(sp)
        STORE   a6, 10 * REGBYTES(sp)
        STORE   a7, 11 * REGBYTES(sp)
        STORE   t3, 12 * REGBYTES(sp)
        STORE   t4, 13 * REGBYTES(sp)
        STORE   t5, 14 * REGBYTES(sp)
        STORE   t6, 15 * REGBYTES(sp)
        //
        // Store the mepc on the stack.
        // This is needed for nestable interrupts.
        //
        csrr    a1, mepc
        STORE   a1, 16 * REGBYTES(sp)
    ENDM

/*********************************************************************
*
*      TRAP_EXIT
*/
TRAP_EXIT MACRO
        //
        // Load the mepc from the stack into the mepc register.
        // This is needed for nestable interrupts.
        //
        LOAD    a1, 16 * REGBYTES(sp)
        csrw    mepc, a1

        // Remain in M-mode after mret
        li      t0, 0x1800
        csrs    mstatus, t0
        //
        // Load caller save registers from stack and return.
        //
        LOAD    ra,  0 * REGBYTES(sp)
        LOAD    t0,  1 * REGBYTES(sp)
        LOAD    t1,  2 * REGBYTES(sp)
        LOAD    t2,  3 * REGBYTES(sp)
        LOAD    a0,  4 * REGBYTES(sp)
        LOAD    a1,  5 * REGBYTES(sp)
        LOAD    a2,  6 * REGBYTES(sp)
        LOAD    a3,  7 * REGBYTES(sp)
        LOAD    a4,  8 * REGBYTES(sp)
        LOAD    a5,  9 * REGBYTES(sp)
        LOAD    a6, 10 * REGBYTES(sp)
        LOAD    a7, 11 * REGBYTES(sp)
        LOAD    t3, 12 * REGBYTES(sp)
        LOAD    t4, 13 * REGBYTES(sp)
        LOAD    t5, 14 * REGBYTES(sp)
        LOAD    t6, 15 * REGBYTES(sp)
        addi    sp, sp, 20 * REGBYTES
        mret
    ENDM

/*********************************************************************
*
*     Publics
*
**********************************************************************
*/
        PUBLIC  vtrap_entry

        PUBWEAK ISR_U_Software
        PUBWEAK ISR_S_Software
        PUBWEAK ISR_U_Timer
        PUBWEAK ISR_S_Timer
        PUBWEAK ISR_U_External
        PUBWEAK ISR_S_External
        PUBWEAK ISR_Local0
        PUBWEAK ISR_Local1
        PUBWEAK ISR_Local2
        PUBWEAK ISR_Local3
        PUBWEAK ISR_Local4
        PUBWEAK ISR_Local5
        PUBWEAK ISR_Local6
        PUBWEAK ISR_Local7
        PUBWEAK ISR_Local8
        PUBWEAK ISR_Local9
        PUBWEAK ISR_Local10
        PUBWEAK ISR_Local11
        PUBWEAK ISR_Local12
        PUBWEAK ISR_Local13
        PUBWEAK ISR_Local14
        PUBWEAK ISR_Local15

/*********************************************************************
*
*     Externals, code
*
**********************************************************************
*/
        EXTERN  OS_TrapHandler
        EXTERN  ISR_M_Software
        EXTERN  ISR_M_Timer
        EXTERN  ISR_M_External

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       Vector table
*/
        SECTION MVECTORS:CODE:NOROOT(2)

        ALIGN 8
vtrap_entry:
        j    USI_Handler   // This vector is ambiguously used for synchronous exceptions as well (see [1])
        j    SSI_Handler
        j    Undefined     // reserved, previously intended for Hypervisor mode
        j    MSI_Handler
        j    UTI_Handler
        j    STI_Handler
        j    Undefined     // reserved, previously intended for Hypervisor mode
        j    MTI_Handler
        j    UEI_Handler
        j    SEI_Handler
        j    Undefined     // reserved, previously intended for Hypervisor mode
        j    MEI_Handler
        j    Undefined     // reserved
        j    Undefined     // reserved
        j    Undefined     // reserved
        j    Undefined     // reserved
        j    LI0_Handler
        j    LI1_Handler
        j    LI2_Handler
        j    LI3_Handler
        j    LI4_Handler
        j    LI5_Handler
        j    LI6_Handler
        j    LI7_Handler
        j    LI8_Handler
        j    LI9_Handler
        j    LI10_Handler
        j    LI11_Handler
        j    LI12_Handler
        j    LI13_Handler
        j    LI14_Handler
        j    LI15_Handler
__Vectors_End


        SECTION MINTERRUPTS:CODE:NOROOT(2)
/*********************************************************************
*
*       USI_Handler
*/
USI_Handler:
        TRAP_ENTRY
    //  call    ISR_U_Software  --- Synchronous exceptions and user mode software interrupts are ambiguously vectored to the same exception handler
        call    OS_TrapHandler  // Implemented in RTOSInit.c
        TRAP_EXIT

/*********************************************************************
*
*       SSI_Handler
*/
SSI_Handler:
        TRAP_ENTRY
        call    ISR_S_Software
        TRAP_EXIT

/*********************************************************************
*
*       MSI_Handler
*/
MSI_Handler:
        TRAP_ENTRY
        call    ISR_M_Software  // Implemented in RTOSInit.c
        TRAP_EXIT

/*********************************************************************
*
*       UTI_Handler
*/
UTI_Handler:
        TRAP_ENTRY
        call    ISR_U_Timer
        TRAP_EXIT

/*********************************************************************
*
*       STI_Handler
*/
STI_Handler:
        TRAP_ENTRY
        call    ISR_S_Timer
        TRAP_EXIT

/*********************************************************************
*
*       MTI_Handler
*/
MTI_Handler:
        TRAP_ENTRY
        call    ISR_M_Timer  // Implemented in RTOSInit.c
        TRAP_EXIT

/*********************************************************************
*
*       UEI_Handler
*/
UEI_Handler:
        TRAP_ENTRY
        call    ISR_U_External
        TRAP_EXIT

/*********************************************************************
*
*       SEI_Handler
*/
SEI_Handler:
        TRAP_ENTRY
        call    ISR_S_External
        TRAP_EXIT

/*********************************************************************
*
*       MEI_Handler
*/
MEI_Handler:
        TRAP_ENTRY
        call    ISR_M_External  // Implemented in RTOSInit.c
        TRAP_EXIT

/*********************************************************************
*
*       LI0_Handler
*/
LI0_Handler:
        TRAP_ENTRY
        call    ISR_Local0
        TRAP_EXIT

/*********************************************************************
*
*       LI1_Handler
*/
LI1_Handler:
        TRAP_ENTRY
        call    ISR_Local1
        TRAP_EXIT

/*********************************************************************
*
*       LI2_Handler
*/
LI2_Handler:
        TRAP_ENTRY
        call    ISR_Local2
        TRAP_EXIT

/*********************************************************************
*
*       LI3_Handler
*/
LI3_Handler:
        TRAP_ENTRY
        call    ISR_Local3
        TRAP_EXIT

/*********************************************************************
*
*       LI4_Handler
*/
LI4_Handler:
        TRAP_ENTRY
        call    ISR_Local4
        TRAP_EXIT

/*********************************************************************
*
*       LI5_Handler
*/
LI5_Handler:
        TRAP_ENTRY
        call    ISR_Local5
        TRAP_EXIT

/*********************************************************************
*
*       LI6_Handler
*/
LI6_Handler:
        TRAP_ENTRY
        call    ISR_Local6
        TRAP_EXIT

/*********************************************************************
*
*       LI7_Handler
*/
LI7_Handler:
        TRAP_ENTRY
        call    ISR_Local7
        TRAP_EXIT

/*********************************************************************
*
*       LI8_Handler
*/
LI8_Handler:
        TRAP_ENTRY
        call    ISR_Local8
        TRAP_EXIT

/*********************************************************************
*
*       LI9_Handler
*/
LI9_Handler:
        TRAP_ENTRY
        call    ISR_Local9
        TRAP_EXIT

/*********************************************************************
*
*       LI10_Handler
*/
LI10_Handler:
        TRAP_ENTRY
        call    ISR_Local10
        TRAP_EXIT

/*********************************************************************
*
*       LI11_Handler
*/
LI11_Handler:
        TRAP_ENTRY
        call    ISR_Local11
        TRAP_EXIT

/*********************************************************************
*
*       LI12_Handler
*/
LI12_Handler:
        TRAP_ENTRY
        call    ISR_Local12
        TRAP_EXIT

/*********************************************************************
*
*       LI13_Handler
*/
LI13_Handler:
        TRAP_ENTRY
        call    ISR_Local13
        TRAP_EXIT

/*********************************************************************
*
*       LI14_Handler
*/
LI14_Handler:
        TRAP_ENTRY
        call    ISR_Local14
        TRAP_EXIT

/*********************************************************************
*
*       LI15_Handler
*/
LI15_Handler:
        TRAP_ENTRY
        call    ISR_Local15
        TRAP_EXIT

/*********************************************************************
*
*       Weak high-level IRQ handler. Can be overwritten in "C".
*/
ISR_U_Software:
ISR_S_Software:
ISR_U_Timer:
ISR_S_Timer:
ISR_U_External:
ISR_S_External:
ISR_Local0:
ISR_Local1:
ISR_Local2:
ISR_Local3:
ISR_Local4:
ISR_Local5:
ISR_Local6:
ISR_Local7:
ISR_Local8:
ISR_Local9:
ISR_Local10:
ISR_Local11:
ISR_Local12:
ISR_Local13:
ISR_Local14:
ISR_Local15:
Undefined:  j       Undefined

/********************************************************************/

        END

/*************************** End of file ****************************/
